# -*- coding: utf-8 -*-
"""
    DWX_AccInfo_API.py - Investor AccountInfo API (Subclass of DWX_API)
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: July 29, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""
import os

from darwinexapis.API.dwx_api import DWX_API

class DWX_AccInfo_API(DWX_API):
    
    def __init__(self,
                 _auth_creds='',
                 _api_url='https://api.darwinex.com',
                 _api_name='investoraccountinfo',
                 _version=2.0,
                 _demo=True):
        
        super(DWX_AccInfo_API, self).__init__(_auth_creds, _api_url,_api_name,_version,_demo)
        
        if _demo:
            print(f'--+--+--\n[KERNEL] InvestorAccountInfoAPI v{_version} initialized in DEMO environment\n--+--+--')
        else:
            print(f'--+--+--\n[KERNEL] InvestorAccountInfoAPI v{_version} initialized in REAL environment\n--+--+--')
        
    ######################################################################### 
    
    def _Get_Accounts_(self):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts', 
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    def _Get_Account_Info_(self, _id=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' + str(_id), 
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Get_Trade_by_ID_(self, _id=0, _tid=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' 
                                   + str(_id)
                                   + '/trades/'
                                   + str(_tid), 
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Get_Order_by_ID_(self, _id=0, _oid=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' 
                                   + str(_id)
                                   + '/orders/'
                                   + str(_oid), 
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Get_Conditional_Order_by_ID_(self, _id=0, _oid=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' 
                                   + str(_id)
                                   + '/conditionalorders/'
                                   + str(_oid), 
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    def _Get_Conditional_Orders_by_Status_(self, 
                                           _id=0, 
                                           _darwin='',
                                           _status='pending',
                                           _page=0,
                                           _perpage=1):
        
        # If valid _status given, then:
        if _status in ['pending','executed','rejected','cancelled','deleted']:
            
            try:
                return self._Call_API_(_endpoint='/investoraccounts/' 
                                       + str(_id)
                                       + '/conditionalorders/'
                                       + str(_status)
                                       + '?productName=' + str(_darwin)
                                       + '&page=' + str(_page)
                                       + '&per_page=' + str(_perpage),
                                       _type='GET', _data='')
            except Exception as ex:
                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
        else:
            print('[ERROR] Invalid Conditional Order Status -> Must be one \
                  of pending, rejected, executed or cancelled.')
            
    #########################################################################
    
    def _Get_Current_Open_Positions_(self, 
                                     _id=0,
                                     _darwin=''):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' 
                                   + str(_id)
                                   + '/currentpositions'
                                   + '?productName=' + str(_darwin),
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    def _Get_Executed_Orders_(self, 
                               _id=0, 
                               _darwin='PLF.4.1',
                               _page=0,
                               _perpage=1):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/' 
                                   + str(_id)
                                   + '/orders/executed'
                                   + '?productName=' + str(_darwin)
                                   + '&page=' + str(_page)
                                   + '&per_page=' + str(_perpage),
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    def _Get_Trades_by_Status_(self, 
                               _id=0, 
                               _darwin='PLF.4.1',
                               _status='open',
                               _page=0,
                               _perpage=1):
        
        # If valid _status given, then:
        if _status in ['open','closed']:
            
            try:
                return self._Call_API_(_endpoint='/investoraccounts/' 
                                       + str(_id)
                                       + '/trades/'
                                       + _status
                                       + '?productName=' + str(_darwin)
                                       + '&page=' + str(_page)
                                       + '&per_page=' + str(_perpage),
                                       _type='GET', _data='')
            except Exception as ex:
                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
        else:
            print('[ERROR] Invalid Trade Status -> Must be one \
                  of open or closed.')
            
    #########################################################################



# -*- coding: utf-8 -*-
"""
    DWX_Info_API.py - DARWIN Info API (Subclass of DWX_API)
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: June 07, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

import os, time

from darwinexapis.MINIONS.dwx_graphics_helpers import DWX_Graphics_Helpers
from darwinexapis.API.dwx_api import DWX_API

import pandas as pd
import numpy as np

pd.set_option('display.width', 1000)
pd.set_option('display.max_columns', 500)

class DWX_Info_API(DWX_API):
    
    def __init__(self,_auth_creds='', _api_url='https://api.darwinex.com', _api_name='darwininfo', _version=1.5, _demo=False):

        super(DWX_Info_API, self).__init__(_auth_creds, _api_url,_api_name, _version, _demo)
        self._graphics = DWX_Graphics_Helpers()

    #########################################################    
    # Function: Get Quotes for all DARWINs in list _symbols #
    #########################################################

    def _Get_Historical_Quotes_(self, _symbols=['THA.4.12','LVS.4.20'], 
                               _start='', #_start=pd.to_datetime('today') - pd.to_timedelta(240, 'D'),
                               _end='', #_end=pd.to_datetime('today'),
                               _endpoint='/products/{}/history/quotes',
                               _plot_title='DWX_Info_API: def _Get_Quotes_() Example',
                               _plot=False,
                               _delay=0.01):
        
        if isinstance(_symbols, list):
            
            _dict = {}
            _count = 1
            
            for darwin in _symbols:
                
                print('\r[DarwinInfoAPI] Getting Quotes for DARWIN {} / {}: ${}'.format(_count, len(_symbols), darwin), 
                      end='', flush=True)
                
                try:
                    
                    # If dates provided, attach query parameters to endpoint
                    if _start not in ['', np.nan]:
                        
                        _ep = _endpoint.format(darwin) + '?start={}&end={}'.format(
                                        int(_start.timestamp())*1000, 
                                        int(_end.timestamp())*1000)
                    
                    else:
                        _ep = _endpoint.format(darwin)
                    
                    # Construct DataFrame from returned JSON
                    _dict[darwin] = pd.DataFrame(data=self._Call_API_(_ep, 'GET', ''))
                            
                    # Assign column names
                    _dict[darwin].columns = ['timestamp', darwin]
                    
                    # Convert ms timestamp to datetime
                    _dict[darwin].timestamp = pd.to_datetime(_dict[darwin].timestamp, unit='ms')
                    
                    # Crop to data before today's date
                    _dict[darwin] = _dict[darwin][_dict[darwin].timestamp.dt.date < pd.to_datetime('today').date()]
                    
                    # Set index to timestamp
                    _dict[darwin] = _dict[darwin].set_index('timestamp')
                    
                    # Sleep
                    if _delay > 0:
                        time.sleep(_delay)
                    
                except Exception as ex:
                    
                    print('[ERROR] Something went wrong while looking up ${}'.format(darwin))
                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
                    continue
                
                # Update counter
                _count += 1
                    
            _retdf = pd.concat([_df for _df in _dict.values() if isinstance(_df, pd.DataFrame)], axis=1)
            _retdf.columns = _symbols
            
            if _plot:
                self._graphics._plotly_dataframe_scatter_(_custom_filename='example_quotes.html', 
                                                          _dir_prefix='MISC/', 
                                                          _df=_retdf,
                                                          _x_title='EOD Timestamp',
                                                          _y_title='DARWIN Quote',
                                                          _main_title=_plot_title)
            
            return _retdf
            
        else:
            print('[ERROR] Please specify symbols as Python list []')    
        
    ######################################################################### 
    
    def _Get_Historical_Scores_(self, 
                                _symbols=['THA.4.12','LVS.4.20'], 
                                _endpoint='/products/{}/history/badges',
                                _plot_title='DWX_Info_API: def _Get_Historical_Scores_() Example',
                                _plot=False,
                                _delay=0.01):
        
        if isinstance(_symbols, list):
            
            _dict = {}
            _count = 1
            
            # Badges dict for later use
            _badge_cols = ['eod_ts','Dp','Ex','Mc','Rs', 
                            'Ra','Os','Cs','Rp','Rm',
                            'Dc','La','Pf','Cp','Ds',
                            'fcal_ts','lcal_ts']
            
            for darwin in _symbols:
                
                print('\r[DarwinInfoAPI] Getting Scores for DARWIN {} / {}: ${}'.format(_count, len(_symbols), darwin),
                      end='', flush=True)
                
                try:
            
                    # Construct DataFrame from returned JSON
                    _dict[darwin] = self._Call_API_(_endpoint.format(darwin), 'GET', '')
                          
                    _dict[darwin] = [_dict[darwin][i][:2] + [x for x in _dict[darwin][i][2]] + _dict[darwin][i][-2:] for i in range(len(_dict[darwin]))]
                    
                    
                    _dict[darwin] = pd.DataFrame(data=_dict[darwin], index=[_dict[darwin][i][0] for i in range(len(_dict[darwin]))])
        
                    # Assign column names
                    _dict[darwin].columns = _badge_cols
                    
                    # Convert ms timestamp to datetime
                    _dict[darwin].eod_ts = pd.to_datetime(_dict[darwin].eod_ts, unit='ms')
                    _dict[darwin].fcal_ts = pd.to_datetime(_dict[darwin].fcal_ts, unit='ms')
                    _dict[darwin].lcal_ts = pd.to_datetime(_dict[darwin].lcal_ts, unit='ms')
                    
                    # Set index to eod_ts
                    _dict[darwin] = _dict[darwin].set_index('eod_ts')
                    
                    # Sleep
                    if _delay > 0:
                        time.sleep(_delay)
                    
                except Exception as ex:
                    
                    print('[ERROR] Something went wrong while looking up ${}'.format(darwin))
                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
                    continue
                
                # Update counter
                _count += 1
                
            # If only one symbol provided, plot scores via Plotly
            if len(_symbols) == 1:
                
                if _plot:
                    self._graphics._plotly_dataframe_scatter_(_custom_filename='example_scores.html', 
                                                              _dir_prefix='MISC/', 
                                                              _df=_dict[_symbols[0]].drop(['fcal_ts','lcal_ts','Ds','Dp'], 
                                                                       axis=1).loc[:, ],
                                                              _x_title='EOD Timestamp',
                                                              _y_title='Score / Investment Attribute',
                                                              _main_title=_plot_title)
                
            return _dict
        
        else:
            print('[ERROR] Please specify symbols as Python list []')    
    
    #########################################################################

    def _Get_DARWIN_Universe_(self, 
                          _status='ALL',
                          _endpoint='/products{}',
                          _query_string='?status={}&page={}&per_page={}',
                          _page=0,
                          _perPage=50,
                          _iterate=True,
                          _delay=0.01):
    
        # Get first batch
        try:
            print('[DarwinInfoAPI] Getting first {} DARWINs..'.format(_perPage))
            _darwins = self._Call_API_(_endpoint \
                                      .format(_query_string \
                                              .format(_status, _page, _perPage)), 
                                              _type='GET',
                                              _data='')
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            return None
            
        if _iterate:
            
            # Calculate number of pages
            _pages = int(_darwins['totalPages'])
            
            # Keep 'content' list of DARWINs, discard everything else
            _darwins = _darwins['content']
            
            print('[DarwinInfoAPI] {} pages of {} DARWINs each found.. iterating, stand by! :muscle:\n'
                  .format(_pages, _perPage))
            
            # Iterate
            for i in range(_page + 1, _pages):
                
                print('\r[DarwinInfoAPI] Getting page {} of {}'.format(i+1, _pages), end='', flush=True)
                
                try:
                    _darwins = _darwins + self._Call_API_(_endpoint \
                                              .format(_query_string \
                                                      .format(_status, i, _perPage)), 
                                                      _type='GET',
                                                      _data='')['content']
                    
                    # Sleep my child.. until next time..
                    if _delay > 0:
                        time.sleep(_delay)
                    
                except Exception as ex:
                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
                    continue
                
        # Return dict
        return pd.DataFrame(_darwins)
        
    ######################################################################### 
        
    """
    Available filters (as of 2019-05-29):
        
        return, drawdown, investors, trades, trader_equity, return_drawdown,
        divergence, days_in_darwinex, current_investment_usd, var, d-score, 
        Ex, Mc, Rs, Ra, Os, Cs, R+, R-, Dc, La, Pf, Sc
        
        For the complete specification, please see:
            --
            https://api.darwinex.com/store/apis/info?name=DarwinInfoAPI
            &version=1.5&provider=admin#/default/filterProductsUsingPOST
            
       
    Example Usage:
    
    # Create API object
    _info = DWX_Info_API()
    
    # 1
    _info._Get_Filtered_DARWINS_(_filters=[['drawdown',-10,0,'6m'],
                                         ['return',5,100,'1m']],
                               _order=['return','12m','DESC'],
                               _page=0,
                               _perPage=50)
    
    # 2
    _info._Get_Filtered_DARWINS_(_filters=[['d-score',50,100,'actual']],
                               _order=['return','12m','DESC'],
                               _page=0,
                               _perPage=50)
    """ 
    
    def _Get_Filtered_DARWINS_(self, 
                               _endpoint='/products',
                               _filters=[['drawdown',-10,0,'6m'],
                                         ['return',5,100,'1m']],
                               _order=['return','12m','DESC'],
                               _page=0,
                               _perPage=50,
                               _delay=0.01):
        
        # Construct filter
        _json = dict(filter=[dict(name=_filters[i][0],
                                  options=[dict(max=_filters[i][2],
                                                min=_filters[i][1],
                                                period=_filters[i][3])]) \
                            for i in range(len(_filters))],
                     order=_order[2],
                     orderField=_order[0],
                     page=_page,
                     perPage=_perPage,
                     period=_order[1])
    
        _rets = []
        
        # Execute
        while _json['page'] != -1:
            
            print('\r[DarwinInfoAPI] Getting page {} of DARWINs that satisfy criteria..' \
                  .format(_json['page']), end='', flush=True)
            
            try:
                _ret = self._Call_API_(_endpoint,
                                      _type='POST',
                                      _data=str(_json).replace('\'', '"'))
                
                if len(_ret) > 0:
                    
                    # Update page number
                    _json['page'] += 1
                    
                    # Add output to running list
                    _rets = _rets + _ret
                    
                    # Sleep between calls
                    time.sleep(_delay)
                else:
                    _json['page'] = -1 # done
                
            except AssertionError:
                print('[ERROR] name, period, min and max lists must be the same length.')
                return None
       
        # Return DataFrame
        return pd.DataFrame(_rets)
    
    ######################################################################### 

    """
    Example Usage:
        
        # from/to dates
        _info._Get_DARWIN_OHLC_Candles_( _symbols=['KVL'], \
                                        _from_dt=str(pd.Timestamp('now') \
                                        - pd.tseries.offsets.BDay(7)), \
                                        _to_dt=pd.Timestamp('now'), \
                                        _resolution='1m')['KVL']
    
        # timeframes
        _info._Get_DARWIN_OHLC_Candles_( _symbols=['KVL'], \
                                        _timeframe='/1D', \
                                        _resolution='1m')['KVL']
                                        
    """
    def _Get_DARWIN_OHLC_Candles_(self, 
                             _symbols=['KVL'],
                             _resolution='1m', # 1m, 5m, 15m, 30m, 1h, 4h, 1d, 1w, 1mn
                             _from_dt='2019-05-31 12:00:00', # UTC
                             _to_dt=str(pd.Timestamp('now')),
                             _timeframe='/1D', # 1D, 1W, 1M, 3M, 6M, 1Y, 2Y, ALL
                             _endpoint='/products/{}/candles{}',
                             _delay=0.01):
    
        # from/to endpoint given priority. change as necessary.
        
        if _from_dt != '':
            
            # Convert human-readable datetimes to millisecond EPOCHs
            _from_epoch = int(pd.Timestamp(_from_dt).timestamp())
            _to_epoch = int(pd.Timestamp(_to_dt).timestamp())
        
            _query_string = f'?resolution={_resolution}&from={_from_epoch}\
                             &to={_to_epoch}'
        
        elif _timeframe != '':
            _query_string = f'{_timeframe}?resolution={_resolution}'
        
        else:
            print('[KERNEL] Inputs not recognized.. please try again.')
            return None
            
        _candles = {}
        
        for _darwin in _symbols:
            
            try:
                print('[DarwinInfoAPI] Getting Candles for {}..'.format(_darwin))
                _d = self._Call_API_(_endpoint \
                                          .format(_darwin, 
                                                  _query_string), 
                                                  _type='GET',
                                                  _data='')
                
                # Parse data into DataFrame
                _candles[_darwin] = pd.DataFrame(
                        data=[_row['candle'] for _row in _d['candles']],
                        index=[_row['timestamp'] for _row in _d['candles']])
                
                # Convert timestamp EPOCHs to human-readable datetimes
                _candles[_darwin].index = pd.to_datetime(_candles[_darwin].index, unit='s')
                
            except Exception as ex:
                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
                return None
        
        return _candles
    
    ######################################################################### 
        



# Let's make some imports:
import os, pprint
import pandas as pd

# Import the different classes:
from darwinexapis.API.TickDataAPI.DWX_TickData_Downloader_API import DWX_TickData_Downloader_API

# Let's create the access token variable:

FTP_CRED = {"username": "your_alpha_username",
            "password": "your_alpha_password",
            "server": "tickdata.darwinex.com"}

# Downloader:
# Try other assets like: GDAXIm, SPN35, XAUUSD... > Watch out with the available dates!
DOWNLOADER = DWX_TickData_Downloader_API(dwx_ftp_user=FTP_CRED['username'], 
                                         dwx_ftp_pass=FTP_CRED['password'],
                                         dwx_ftp_hostname=FTP_CRED['server'],
                                         dwx_ftp_port=FTP_CRED['port'])

# Create the path you wish to save the data:                                         
path_to_save = os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/')

###################################################

# One hour data (be sure to put the hours with two characters > 0 == 00, 5 == 05, 23 ==23):
bid_hour_data = DOWNLOADER._download_one_hour_data_bid(_asset='WS30', 
                                                       _date='2018-10-01', 
                                                       _hour='22',
                                                       _verbose=True)
DOWNLOADER._save_df_to_csv(bid_hour_data, which_path=path_to_save)                                                       
DOWNLOADER._save_df_to_pickle(bid_hour_data, which_path=path_to_save)

ask_hour_data = DOWNLOADER._download_one_hour_data_ask(_asset='WS30', 
                                                       _date='2018-10-01', 
                                                       _hour='22',
                                                       _verbose=True)
DOWNLOADER._save_df_to_csv(ask_hour_data, which_path=path_to_save)                                                       
DOWNLOADER._save_df_to_pickle(ask_hour_data, which_path=path_to_save)

###################################################

# One day data (be sure to also put the dates with two characters):
bid_day_data = DOWNLOADER._download_one_day_data_bid(_asset='WS30', 
                                                     _date='2018-10-01',
                                                     _verbose=True)
DOWNLOADER._save_df_to_csv(bid_day_data, which_path=path_to_save)                                                     
DOWNLOADER._save_df_to_pickle(bid_day_data, which_path=path_to_save)                                                     

ask_day_data = DOWNLOADER._download_one_day_data_ask(_asset='WS30', 
                                                     _date='2018-10-01',
                                                     _verbose=True)
DOWNLOADER._save_df_to_csv(ask_day_data, which_path=path_to_save)                                                     
DOWNLOADER._save_df_to_pickle(ask_day_data, which_path=path_to_save)                                                     

###################################################

# Between two dates data:
bid_date_data = DOWNLOADER._download_month_data_bid(_asset='WS30', 
                                                    _start_date='2018-10-01', 
                                                    _end_date='2018-10-04', 
                                                    _verbose=True)
DOWNLOADER._save_df_to_csv(bid_date_data, which_path=path_to_save)                                                    
DOWNLOADER._save_df_to_pickle(bid_date_data, which_path=path_to_save)                                                    

ask_date_data = DOWNLOADER._download_month_data_ask(_asset='WS30', 
                                                    _start_date='2018-10-01', 
                                                    _end_date='2018-10-04', 
                                                    _verbose=True)
DOWNLOADER._save_df_to_csv(ask_date_data, which_path=path_to_save)                                                    
DOWNLOADER._save_df_to_pickle(ask_date_data, which_path=path_to_save)                                                    



# -*- coding: utf-8 -*-
"""
    Sample code:
    DWX_TickData_Reader_API.py
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

# Do some imports:
from pathlib import Path
import pandas as pd
import numpy as np
import gzip, os
import matplotlib.pyplot as plt

pd.set_option('display.max_columns', 500)
pd.set_option('display.width', 1000)

class DWX_TickData_Reader_API():
    
    '''This class will read the bid and ask files on a directory and create
    a dataframe with both price sequences (i.e. time series).'''

    def __init__(self, _bids_file='<INSERT_PATH>', _asks_file='<INSERT_PATH>'):
        
        # Get the symbol:
        self._file_name_csv = _bids_file.split('/')[-1].replace('BID', 'BID_ASK').replace('pkl', 'csv')
        self._asset_name = _bids_file.split('_')[0]
        
        # Fix some variables:
        self._bid_price_col_name = f'{self._asset_name}_bid_price'
        self._ask_price_col_name = f'{self._asset_name}_ask_price'
        self._bid_size_col_name = f'{self._asset_name}_bid_size'
        self._ask_size_col_name = f'{self._asset_name}_ask_size'

        # Define the variables:
        self._bids_file = _bids_file
        self._asks_file = _asks_file
    
    def _construct_data_(self, _filename):
        
        _df = pd.read_pickle(_filename)
        
        if 'BID' in _filename:
            _df.columns = [self._bid_price_col_name,self._bid_size_col_name]
        elif 'ASK' in _filename:
            _df.columns = [self._ask_price_col_name,self._ask_size_col_name]
            
        _df.index.name = f'{self._asset_name}_timestamp'
        
        return _df.apply(pd.to_numeric)
    
    def _get_symbol_as_dataframe_(self, _convert_epochs=True,
                                        _check_integrity=False,
                                        _calc_spread = False,
                                        _reindex=[],
                                        _precision='tick',
                                        _daily_start=22,
                                        _symbol_digits=5):
        
        """
        See http://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html
        for .resample() rule / frequency strings.        
        """
    
        # BIDS
        _bids = self._construct_data_(self._bids_file)
        
        # ASKS
        _asks = self._construct_data_(self._asks_file)
                
        _df = _asks.merge(_bids, how='outer', left_index=True, right_index=True, copy=False).fillna(method='ffill').dropna()
            
        # Calculate spread?
        if _calc_spread:
            _df[f'{self._asset_name}_spread'] = abs(np.diff(_df[[self._ask_price_col_name,self._bid_price_col_name]]))
        
        # Convert timestamps?
        if _convert_epochs:
            _df.index = pd.to_datetime(_df.index, unit='ms')
        
        # Reindex to selected columns?
        if len(_reindex) > 0:
            _df = _df.reindex(_reindex, axis=1)
            
        # Resample?
        if _precision != 'tick':
            _df[f'{self._asset_name}_mid_price'] = round((_df[self._ask_price_col_name] + _df[self._bid_price_col_name]) / 2, _symbol_digits)
            
            if _precision not in ['B','C','D','W','24H']:
                _df = _df[f'{self._asset_name}_mid_price'].resample(rule=_precision).ohlc()
            else:
                _df = _df[f'{self._asset_name}_mid_price'].resample(rule=_precision, base=_daily_start).ohlc().dropna()
            
        # Check data integrity?
        if _check_integrity:
            
            print('\n\n[INFO] Checking data integrity..')
            self._integrity_check_(_df)
        
        return _df
    
    def _save_df_to_csv(self, dataframe_to_save, which_path=None):

        # Save:
        if which_path:
            # It will save the data to the specified path:
            dataframe_to_save.to_csv(which_path + self._file_name_csv)
        else:
            # It will save the data in the working directory:
            dataframe_to_save.to_csv(self._file_name_csv) 

    def _integrity_check_(self, _df):
        
        if isinstance(_df, pd.DataFrame) == False:
            
            print('[ERROR] Input must be a Pandas DataFrame')
            
        else:
            
            _diff = _df.index.to_series().diff()
            
            print('\n[TEST #1] Data Frequency Statistics\n--')
            print(_diff.describe())
            
            print('\n[TEST #2] Mode of Gap Distribution\n--')
            print(_diff.value_counts().head(1))
            
            print('\n[TEST #3] Hourly Spread Distribution\n--')
            _df.groupby(_df.index.hour)[f'{self._asset_name}_mid_price'].mean().plot(
                    xticks=range(0,24), 
                    title='Average Spread by Hour (UTC)')

            # Plot it:
            plt.show()


# APIs:
from .API.InfoAPI.DWX_Info_API import DWX_Info_API

from .API.InvestorAccountInfoAPI.DWX_AccInfo_API import DWX_AccInfo_API

from .API.QuotesAPI.DWX_Quotes_API import DWX_Quotes_API

from .API.TradingAPI.DWX_Trading_API import DWX_Trading_API

from .API.WebSocketAPI.DWX_WebSocket_API import DWX_WebSocket_API

from .API.TickDataAPI.DWX_TickData_Downloader_API import DWX_TickData_Downloader_API
from .API.TickDataAPI.DWX_TickData_Reader_API import DWX_TickData_Reader_API

from .API.DarwinDataAnalyticsAPI.DWX_Data_Analytics_API import DWX_Darwin_Data_Analytics_API

# Minions:
from .MINIONS.dwx_graphics_helpers import DWX_Graphics_Helpers


# Let's make some imports:
import os, pprint
import pandas as pd

# Import the different classes:
from darwinexapis.API.TickDataAPI.DWX_TickData_Reader_API import DWX_TickData_Reader_API

# Reader:
bid_file_pkl = os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/WS30_BID_2018-10-01_23.pkl')
ask_file_pkl = os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/WS30_ASK_2018-10-01_23.pkl')
path_to_save = os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/')

# Generate the object: 
READER = DWX_TickData_Reader_API(_bids_file=bid_file_pkl, 
                                 _asks_file=ask_file_pkl)

# Generate the dataframe: 
readed_dataframe = READER._get_symbol_as_dataframe_(_convert_epochs=True,
                                                    _check_integrity=True,
                                                    _calc_spread=True,
                                                    _reindex=['ask_price', 'bid_price', 'spread'],
                                                    _precision='tick')

# Save it:                                                                                                    
READER._save_df_to_csv(readed_dataframe, which_path=path_to_save)


# -*- coding: utf-8 -*-
"""
    dwx_analytics.py - Pythonic access to raw DARWIN analytics data via FTP
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: October 17, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""
import gzip
import json, os
import pandas as pd
from tqdm import tqdm
from ftplib import FTP
from io import BytesIO
from matplotlib import pyplot as plt

class DWX_Darwin_Data_Analytics_API():
    
    '''This API has the ability to download DARWIN data and analyze it.'''

    def __init__(self, dwx_ftp_user, dwx_ftp_pass, dwx_ftp_hostname, dwx_ftp_port):
        
        """Initialize variables, setup byte buffer and FTP connection.
        
        Parameters
        ----------
        ftp_server : str
            FTP server that houses raw DARWIN data
            
        ftp_username : str
            Your Darwinex username
            
        ftp_password : str
            Your FTP password (NOT your Darwinex password)
            
        ftp_port : int
            Port to connect to FTP server on.
        --
        """
        
        # Analytics Headers
        self.analytics_headers = {'AVG_LEVERAGE': ['timestamp','periods','darwin_vs_eurusd_volatility'],
                                  'ORDER_DIVERGENCE': ['timestamp','instrument','usd_volume','latency','divergence'],
                                  'RETURN_DIVERGENCE': ['timestamp','quote','quote_after_avg_divergence'],
                                  'MONTHLY_DIVERGENCE': ['timestamp','average_divergence','monthly_divergence'],
                                  'DAILY_FIXED_DIVERGENCE': ['timestamp','profit_difference'],
                                  'DAILY_REAL_DIVERGENCE': ['timestamp','profit_difference']}
    
        # Setup data container
        self.retbuf = BytesIO()
        
        # Setup data access mode (file or FTP)
        self.mode = 0 # Default is file.
        
        try:
            self.server = FTP(dwx_ftp_hostname)
            self.server.login(dwx_ftp_user, dwx_ftp_pass)
            
            # 200+ codes signify success.
            if str(self.server.lastresp).startswith('2'):
                print('[KERNEL] FTP Connection Successful. Data will now be pulled from Darwinex FTP Server.')
                self.mode = 1 # 1 = FTP, 0
            
            print(f'[KERNEL] Last FTP Status Code: {self.server.lastresp} | Please consult https://en.wikipedia.org/wiki/List_of_FTP_server_return_codes for code definitions.')
                
        except Exception as ex:
            print(f'Exception: {ex}')
            exit(-1)
                
    ##########################################################################
    
    """Parse a line containing a list. Only works for max one list or one list of lists."""
    
    def parse_line(self, line):
        
        for start, end in [['[[', ']]'], ['[', ']']]:
            if start in line:
                ls = line.split(start)
                ls1 = ls[1].split(end)
                return ls[0].split(',')[:-1] + [json.loads(start+ls1[0].replace("'", '"')+end)] + ls1[1].split(',')[1:]
        return line.split(',')
    
    def get_data_from_ftp(self, darwin, data_type):
        
        """Connect to FTP server and download requested data for DARWIN.
        
        For example, darwin='PLF' and data_type='AVG_LEVERAGE' results in this
        code retrieving the file 'PLF/AVG_LEVERAGE' from the FTP server.
        
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        data_type : str
            Must be a key in self.analytics_headers dictionary.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Clear / reinitialize buffer
        self.retbuf = BytesIO()
        
        self.server.retrbinary(f"RETR {darwin}/{data_type}", self.retbuf.write)
        self.retbuf.seek(0)
        
        # Extract data from BytesIO object
        ret = []
        
        while True:
            line = self.retbuf.readline()
            if len(line) > 1:
                ret.append(self.parse_line(line.strip().decode()))
            else:
                break
        
        # Return as Dataframe
        return pd.DataFrame(ret)
    
    def get_data_from_file(self, darwin, data_type):
        
        """Read data from local file stored in path darwin/filename
        
        For example, darwin='PLF' and data_type='AVG_LEVERAGE' results in this
        code retrieving the file 'PLF/AVG_LEVERAGE' from the current directory.
        
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        data_type : str
            Must be a key in self.analytics_headers dictionary.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        if self.mode == 0:
            print(f'Retrieving data from file for DARWIN {darwin}...')
            return pd.read_csv(f'{str(darwin).upper()}/{str(data_type).upper()}', header=None)
        else:
            print(f'Retrieving data from FTP Server for DARWIN {darwin}...')
            return self.get_data_from_ftp(str(darwin).upper(), str(data_type).upper())
    
    def save_data_to_csv(self, dataframe_to_save, which_path, filename):

        # Save:
        if which_path:
            # It will save the data to the specified path:
            dataframe_to_save.to_csv(which_path + filename + '.csv')
        else:
            # It will save the data in the working directory:
            dataframe_to_save.to_csv(filename + '.csv')

    ##########################################################################
    
    def get_analytics(self, darwin, data_type):
        
        """Get, index and prepare requested data.
        
        For example, darwin='PLF' and data_type='AVG_LEVERAGE' results in:
            
            - the code retrieving the file 'PLF/AVG_LEVERAGE'
            - converting millisecond timestamps column to Pandas datetime
            - Setting the above converted timestamps as the index
            - Dropping the timestamp column itself.            
        
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        data_type : str
            Must be a key in self.analytics_headers dictionary.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        df = self.get_data_from_file(darwin, data_type)
        
        df.columns = self.analytics_headers[data_type]
        df.set_index(pd.to_datetime(df['timestamp'], unit='ms'), inplace=True)
        df.drop(['timestamp'], axis=1, inplace=True)
        
        return df
    
    ##########################################################################
        
    def get_darwin_vs_eurusd_volatility(self, darwin, plot=True):
        
        """Get the evolution of the given DARWIN's volatility vs that of the EUR/USD.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        plot : bool
            If true, produce a chart as defined in the method.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'AVG_LEVERAGE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type)
            
        # DARWIN vs EURUSD volatility is a list. We need the last value
        df.loc[:,self.analytics_headers[data_type][-1]] = \
            df.loc[:,self.analytics_headers[data_type][-1]].apply(eval).apply(lambda x: x[-1])
            
        if plot:
            df['darwin_vs_eurusd_volatility'].plot(title=f'${darwin}: DARWIN vs EUR/USD Volatility',
                                                   figsize=(10,8))
            
        # Return processed data
        return df
        
    ##############################################################################
        
    def get_order_divergence(self, darwin,
                             plot=True):
        
        """Get the evolution of the given DARWIN's replication latency and investor
        divergence, per order executed by the trader.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        plot : bool
            If true, produce a chart as defined in the method.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'ORDER_DIVERGENCE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type)
        
        # Convert values to numeric
        df[['latency','usd_volume','divergence']] = df[['latency','usd_volume','divergence']].apply(pd.to_numeric, errors='coerce')
        
        # Plot
        if plot:
            
            fig = plt.figure(figsize=(10,12))

            # 2x1 grid, first plot
            ax1 = fig.add_subplot(211)
            ax1.xaxis.set_label_text('Replication Latency (ms)')
            
            # 2x1 grid, second plot
            ax2 = fig.add_subplot(212)
            ax2.xaxis.set_label_text('Investor Divergence')
            
            # Plot Median Replication Latency by Instrument
            df.groupby('instrument').latency.median()\
                .sort_values(ascending=True).plot(kind='barh',\
                             title=f'${darwin} | Median Order Replication Latency (ms)',\
                             ax=ax1)
            
            # Plot Median Investor Divergence by Instrument
            df.groupby('instrument').divergence.median()\
                .sort_values(ascending=True).plot(kind='barh',\
                             title=f'${darwin} | Median Investor Divergence per Order',\
                             ax=ax2)
                
            fig.subplots_adjust(hspace=0.2)
                
        # Return processed data
        return df.dropna()
    
    ##########################################################################
    
    def get_return_divergence(self, darwin, plot=True):
        
        """Get the evolution of the given DARWIN's Quote and Quote after applying
        average investors' divergence.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        plot : bool
            If true, produce a chart as defined in the method.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'RETURN_DIVERGENCE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type).apply(pd.to_numeric, errors='coerce')
        
        if plot:
            df.plot(title=f'${darwin} | Quote vs Quote with Average Divergence',
                    figsize=(10,8))
            
        return df
    
    ##########################################################################
    
    def get_monthly_divergence(self, darwin):
        
        """Get the evolution of the given DARWIN's average and monthly divergence.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'MONTHLY_DIVERGENCE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type).apply(pd.to_numeric, errors='coerce')
        
        return df
    
    ##########################################################################
    
    def get_daily_fixed_divergence(self, darwin, plot=True):
        
        """Analyses the effect of applying a fixed divergence (10e-5) on the profit.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        plot : bool
            If true, produce a chart as defined in the method.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'DAILY_FIXED_DIVERGENCE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type).apply(pd.to_numeric, errors='coerce')
        
        if plot:
            df.plot(title=f'${darwin} | Effect of 10e-5 Fixed Divergence on profit',
                    figsize=(10,8))
            
        return df
    
    ##########################################################################
    
    def get_daily_real_divergence(self, darwin, plot=True):
        
        """Analyse the effect of applying the investors' divergence on the profit.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        plot : bool
            If true, produce a chart as defined in the method.
            
        Returns
        -------
        df
            Pandas DataFrame
        --
        """
        
        # Set required data type
        data_type = 'DAILY_REAL_DIVERGENCE'
        
        # Get raw data into pandas dataframe
        df = self.get_analytics(darwin, data_type).apply(pd.to_numeric, errors='coerce')
        
        if plot:
            df.plot(title=f'${darwin} | Effect of Investor Divergence on profit',
                    figsize=(10,8))
            
        return df
    
    ##########################################################################
    
    def get_quotes_from_ftp(self, 
                            darwin='PLF',
                            suffix='4.1',
                            monthly=True, # If set to False, month/year used.
                            month='01',
                            year='2019'):
        
        """Download Quote data for any DARWIN directly via FTP.
                
        Parameters
        ----------
        darwin : str
            DARWIN ticker symbol, e.g. $PLF
            
        suffix : str
            Reflects risk, '4.1' being for 10% VaR DARWIN assets.
            
        monthly : bool
            If set to True, month and year arguments are ignored, and ALL 
            data available for the DARWIN is downloaded.
            
        month : str
            Data on the FTP server has the following directory structure:    
                DARWIN_Symbol -> {year}-{month} -> *.csv.gz files
                e.g. PLF/2019-01/PLF....csv.gz
                
                Specifies month for {year}-{month} tuple as above.
        
        year : str
            Data on the FTP server has the following directory structure:    
                DARWIN_Symbol -> {year}-{month} -> *.csv.gz files
                e.g. PLF/2019-01/PLF....csv.gz
                
                Specifies year for {year}-{month} tuple as above.
            
        Returns
        -------
        df
            Pandas DataFrame containing Quotes, indexed by timeframe.
        --
        """
        quote_files = []
        roots = []
        
        if monthly:
            
            tqdm.write(f'\n[KERNEL] Searching for Quote data for DARWIN {darwin}, please wait..', end='')
            self.server.retrlines(f'NLST {darwin}/quotes/', roots.append)
            
            roots_pbar = tqdm(roots, position=0, leave=True)
            for root in roots_pbar:
                try:
                    roots_pbar.set_description("Getting filenames for month: %s" % root)
                    root_files = []
                
                    self.server.retrlines(f'NLST {darwin}/quotes/{root}', root_files.append)
                
                    # Finalize filenames
                    quote_files += [f'{darwin}/quotes/{root}/{root_file}'\
                                    for root_file in root_files if '{}.{}'.format(darwin, suffix) in root_file]
                except Exception as ex:
                    print(ex)
                    return
            
        elif pd.to_numeric(month) > 0 and pd.to_numeric(year) > 2010:
            
            tqdm.write(f'\n[KERNEL] Asserting data for month {year}-{month}, please wait..', end='')    
            quote_files = []
            
            try:
                self.server.retrlines(f'NLST {darwin}/quotes/{year}-{month}/', quote_files.append)
                quote_files = [f'{darwin}/quotes/{year}-{month}/{quote_file}'\
                                    for quote_file in quote_files if '{}.{}'.format(darwin, suffix) in quote_file]
            except Exception as ex:
                print(ex)
                return
        else:
            print('\n[ERROR] Please either set monthly=True or ensure both month and year have integer values')
            return
            
        # Process tick data files
        tqdm.write(f'\n[KERNEL] {len(quote_files)} files retrieved.. post-processing now, please wait..', end='')
        ticks_df = pd.DataFrame()
        ticks_pbar = tqdm(quote_files, position=0, leave=True)
        
        for tick_file in ticks_pbar:
            
            try:
                ticks_pbar.set_description("Processing %s" % tick_file)
                
                # Clear / reinitialize buffer
                self.retbuf = BytesIO()
            
                self.server.retrbinary(f"RETR {tick_file}", self.retbuf.write)
                self.retbuf.seek(0)
                
                # Extract data from BytesIO object
                ret = [line.strip().decode().split(',') for line in gzip.open(self.retbuf)]
                ticks_df = pd.concat([ticks_df, pd.DataFrame(ret[1:])], axis=0)
                
            except Exception as ex:
                print(ex)
        
        # Clean up
        ticks_df.columns = ['timestamp','quote']
        ticks_df.timestamp = ticks_df.timestamp.apply(pd.to_numeric)
        ticks_df.set_index('timestamp', drop=True, inplace=True)
        ticks_df.index = pd.to_datetime(ticks_df.index, unit='ms')
        ticks_df.quote = ticks_df.quote.apply(pd.to_numeric)
        
        # Return DataFrame
        return ticks_df.dropna()
    
    ##########################################################################



# -*- coding: utf-8 -*-
"""
    DWX WebSocket API - Subclass of DWX_API for Quotes Streaming
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: June 25, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

# Do some imports:
import os, time, sys

from darwinexapis.API.dwx_api import DWX_API
import websockets, json, asyncio

class DWX_WebSocket_API(DWX_API):
    
    def __init__(self, _auth_creds='', _api_url='ws://api.darwinex.com/quotewebsocket/1.0.0', _api_name='', _version=0.0):
        
        # Set to WebSocket URL and other variables:
        self._api_url = _api_url
        self._auth_creds = _auth_creds
        self._api_name = _api_name
        self._version = _version

        # Initialize the DWX_API class:
        super(DWX_WebSocket_API, self).__init__(self._auth_creds, self._api_url, self._api_name, self._version)
        
        # If false, stop polling data from websocket
        self._active = True
        self._websocket = None
    
    async def subscribe(self, _symbols=['DWZ.4.7','DWC.4.20','LVS.4.20','SYO.4.24','YZZ.4.20']):
        
        #print(f'[SUBSCRIBE_UP] - AUTH_HEADERS: {self._auth_headers}')

        async with websockets.connect(self._api_url, extra_headers=self._auth_headers) as websocket:

           # Subscribe to symbols
           await websocket.send(json.dumps({ 'op': 'subscribe', 'productNames' :_symbols}))
           
           # If _active is True, process data received.
           while self._active:
               
               # If the time is greater that the time + the expires in > issue refresh:s
                if time.time() > self.AUTHENTICATION.expires_in:

                    print('\n[SUBSCRIBE] - The expiration time has REACHED > ¡Generate TOKENS!')
                    # Generate new token:
                    self.AUTHENTICATION._get_access_refresh_tokens_wrapper()

                    # Re-run the loop:
                    print('[SUBSCRIBE] - Need to re-run the loop with new TOKENS...')
                    return

                else:
                    print('\n[SUBSCRIBE] - The expiration time has NOT reached yet > Continue...')

                # Keep returning:
                _ret = await websocket.recv()
               
                # Insert your Quote handling logic here
                # {"op":"hb","timestamp":1587838905842} > Heartbeats.
                # Reference: https://api.darwinex.com/store/site/pages/doc-viewer.jag?docName=Product%20Quotes%20WebSocket%20subscription%20walkthrough&name=QuoteWebSocket&version=1.0.0&provider=admin&
                print(_ret)

    def run(self, _symbols=['DWZ.4.7','DWC.4.20','LVS.4.20','SYO.4.24','YZZ.4.20']):
        
        self._symbols = _symbols
        self.event_loop = asyncio.get_event_loop()
        
        try:
            self.event_loop.run_until_complete(self.subscribe(_symbols))

        except RuntimeError as ex:

            print(f'[RUNTIME ERROR] > {ex}')
            
        except KeyboardInterrupt:

            print('[EXCEPTION] > ¡KeyboardInterrupt Exception!')

        else:

            # Re-run the loop and stop it:
            print('[RUN_ELSE] - Tokens generated > We will re-run the loop and start the WS connection again')

            # Cancel all tasks:
            for task in asyncio.Task.all_tasks():
                task.cancel()
            print('[RUN_ELSE] - All tasks cancelled')

            # Stop the loop to clean:
            self.stop_and_close()

    def stop_and_close(self):
        
        """Stop and close loop"""

        self.event_loop.stop()
        self.event_loop.close()
        print('[CLOSE] - Loop stopped and closed')

        # Launch:
        self.launch_loop_again()

    def launch_loop_again(self):

        # Execute again the DWX_API init:
        super(DWX_WebSocket_API, self).__init__(self._auth_creds, self._api_url, self._api_name, self._version)

        # Create it and assign it:
        asyncio.set_event_loop(asyncio.new_event_loop())
        print('[LAUNCH] - New event loop set and assigned > Will run the coroutine again..')
        self.run(_symbols=self._symbols)



# -*- coding: utf-8 -*-
"""
    DWX Graphics - Helper Class
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: June 20, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""
import plotly.offline as po
import plotly.graph_objs as go
import matplotlib.ticker as ticker

class DWX_Graphics_Helpers():
    
    def __init__(self):
        pass
    
    ##########################################################################
        
    def _mpl_plot_axis_(self, _plt, _ax, _df, _darwin, 
                          _x_title, _y_title, _line_color,
                          _line_width, _tail_max,
                          _bgcolor, _tfont):
        
        # Matplotlib y-axis formatter
        @ticker.FuncFormatter
        def major_formatter(x, pos):
            return "%.2f" % x
        
        _ax.cla()
        _ax.set_title(_darwin, **_tfont)
        _ax.set_facecolor(_bgcolor)
            
        _ax.ticklabel_format(useOffset=False)
        _ax.yaxis.set_major_formatter(major_formatter)
        _ax.yaxis.set_major_formatter(major_formatter)

        _ax.set_xticklabels([])
        _ax.set_xlabel(_x_title, **_tfont)
        _ax.set_ylabel(_y_title, **_tfont)
        
        _ax.set_ylim(bottom=_df[_darwin].tail(_tail_max).min() - 0.01,
                                              top=_df[_darwin].tail(_tail_max).max() + 0.01)
                
        _ax.plot(_df.tail(_tail_max).index, 
                  _df[_darwin].tail(100).values,
                  color=_line_color,
                  linewidth=_line_width)
                  #marker='.')
        
        _ax.grid(linestyle='-', linewidth='0.5', color='grey')
        
        _plt.tight_layout() 
        _plt.pause(0.01)
        
    
    ##########################################################################
    
    def _generate_scatter_single_(self, _df, _name):
        return go.Scatter(
                        x = _df.index,
                        y = _df,
                        name = _name)
        
    def _generate_scatter_list_(self, _df):
        
        return [self._generate_scatter_single_(
                        _df = _df[_darwin],
                        _name = _darwin
                    ) for _darwin in _df.columns]
        
        """
        return [go.Scatter(
                        x = _df.index,
                        y = _df[_darwin],
                        name = _darwin) 
                    for _darwin in _df.columns]
        """
        
    ##########################################################################
    
    
    ##########################################################################
    
    def _plotly_multi_scatter_(self, 
                                  _data=None,
                                  _updatemenus=None,
                                  _title=None,
                                  _legend=None,
                                  _dir_prefix='../MISC/',
                                  _x_title='Date / Time',
                                  _y_title='Quote'):
        
        # Setup Layout
        _layout = go.Layout(
            title=_title,
            showlegend=_legend,
            updatemenus=_updatemenus,
            xaxis=dict(                
                title=_x_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#ffffff'
                )
            ),
            yaxis=dict(
                title=_y_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#ffffff'
                )
            ),
            hoverlabel = dict(namelength = -1),
            images=[dict(
                #source="https://d139oolcsxoepg.cloudfront.net/2_108_0/images/logos/darwinex/logo-darwinex.svg",
                source="https://avatars2.githubusercontent.com/u/26509507?s=460&v=4",
                xref="paper", yref="paper",
                x=1.04, y=0.99,
                sizex=0.105, sizey=0.105,
                xanchor="left", yanchor="bottom"
              )]
        )
            
        fig = go.Figure(data=_data, layout=_layout)
        po.plot(fig, filename=_dir_prefix + f'{_title.replace(" ", "-")}.html')
    
    ##########################################################################
        
    # Single Series Scatter Chart
    def _plotly_dataframe_scatter_(self, 
                        _custom_filename='',
                        _dir_prefix='../MISC/',
                        _df=None,
                        _name='',
                        _x_title='Date / Time', 
                        _y_title='Quote', 
                        _main_title='DWX_Info_API: def _Get_Quotes_() Example',
                        _annotations=[],
                        _plot_only=False):
        
        if _df is not None:
            
            _data = self._generate_scatter_list_(_df)
            
        else:
            print('[ERROR] Can\'t plot thin air I\'m afraid... :)')
            return None
    
        # Setup Layout
        _layout = go.Layout(
            title=_main_title,
            xaxis=dict(
                title=_x_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#7f7f7f'
                )
            ),
            yaxis=dict(
                title=_y_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#7f7f7f'
                )
            ),
            hoverlabel = dict(namelength = -1),
            annotations=_annotations
        )
                
        # Create figure
        fig = go.Figure(data=_data, layout=_layout)
        
        if _plot_only:
            po.iplot(fig)
        else:
            # Set output filename and plot
            if _custom_filename != '':
                po.plot(fig, filename=_dir_prefix + _custom_filename)
            else:
                po.plot(fig, filename=_dir_prefix + _main_title + '.html')
        
    ##########################################################################
    
    def _plotly_scatter_y2(self, 
                           _custom_filename='',
                           _t1_data=None, _t2_data=None,
                           _x1=None, _y1=None, _x1_title=None, 
                           _y1_range=None, _y1_title=None,_x2=None, 
                           _y2=None, _x2_title=None, _y2_range=None, 
                           _y2_title=None, _main_title=None,
                           _annotations=[]):
        
        if _t1_data is None or _t2_data is None:
            print('[ÈRROR] You must provide lists of Scatter objects to proceed.')
            return None
        
        _data = _t1_data + _t2_data
        
        _layout = go.Layout(
                
            title=_main_title,
            xaxis=dict(
                title=_x1_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#7f7f7f'
                )
            ),
             
            yaxis=dict(
                title=_y1_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#7f7f7f'
                )
                # overlaying='y',
                # side='left'
            ),
                
            yaxis2=dict(
                title=_y2_title,
                titlefont=dict(
                    family='Courier New, monospace',
                    size=18,
                    color='#7f7f7f'
                ),
                overlaying='y',
                #range=_y2_range,
                side='right'
            ),
            hoverlabel = dict(namelength = -1),
            annotations=_annotations
        )
                
        fig = go.Figure(data=_data, layout=_layout)
        
        if _custom_filename != '':
            po.plot(fig, filename=_custom_filename)
        else:
            po.plot(fig, filename=_main_title + '.html')
            
        #po.plot(_fig, filename=self._reports_dir_prefix + _main_title + '.html')
        
    #########################################################################



# Do some imports:
import os, pandas as pd

# Import the class:
from darwinexapis.API.DarwinDataAnalyticsAPI.DWX_Data_Analytics_API import DWX_Darwin_Data_Analytics_API

# Create the object:
FTP_CRED = {"username": "your_alpha_username",
            "password": "your_alpha_password",
            "server": "darwindata.darwinex.com",
            "port": 21}

ANALYZER = DWX_Darwin_Data_Analytics_API(dwx_ftp_user=FTP_CRED['username'], 
                                         dwx_ftp_pass=FTP_CRED['password'],
                                         dwx_ftp_hostname=FTP_CRED['server'],
                                         dwx_ftp_port=FTP_CRED['port'])

# Download data of certain analytics variable:
dataFrameReturned = ANALYZER.get_analytics(darwin='LVS', data_type='RETURN_DIVERGENCE')
ANALYZER.save_data_to_csv(dataFrameReturned, which_path=os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/'), filename='LVS_AVG_LEVERAGE')
print(dataFrameReturned)

# Get quote date for DARWINs:
# This call will get all the data and will take some time to execute.
quotes = ANALYZER.get_quotes_from_ftp(darwin='PLF',
                                      suffix='4.1',
                                      monthly=False, # If set to False, month/year used > If True ALL data available
                                      month='01',
                                      year='2019')
ANALYZER.save_data_to_csv(quotes, which_path=os.path.expandvars('${HOME}/Desktop/darwinexapis/darwinexapis/EXAMPLE_DATA/'), filename='LVS_Quotes')
print(quotes.head())                                      


# -*- coding: utf-8 -*-
"""
    DWX_API - Superclass for all sub-APIs
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: June 29, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""

# Do some imports:
import os, requests, json, time
from darwinexapis.API.DWX_API_Auth import DWX_API_AUTHENTICATION

class Decorators():

    @staticmethod
    def refreshTokenDecorator_APIs(func_to_be_decorated):

        '''Example use > put the decorator upfront the method that will make request:

        @refreshTokenDecorator
        def someRequest(self):
            # make our API request
            pass'''

        # The function that is used to check the TOKEN and refresh if necessary
        def wrapper_of_the_func(self, *args, **kwargs):

            # If the time is greater that the time + the expires in > issue refresh:
            if time.time() > self.AUTHENTICATION.expires_in:

                print('\n[DECORATOR] - The expiration time has REACHED > ¡Generate TOKENS!')
                # Generate new credentials:
                self.AUTHENTICATION._get_access_refresh_tokens_wrapper()

                # Call the function after getting new creds:
                # NOTE: no need to call the function here > The decorator will call it after.
                #self._Call_API_(*args, **kwargs)
                
            else:
                print('\n[DECORATOR] - The expiration time has NOT reached yet > Continue...')
            return func_to_be_decorated(self, *args,**kwargs)

        # Return the decorated function > Something like: <function refreshTokenDecorator.<locals>.wrapper at 0x7f3c5dfd42f0>
        return wrapper_of_the_func

class DWX_API(object):
    
    def __init__(self,
                 _auth_creds='',
                 _api_url='https://api.darwinex.com',
                 _api_name='darwininfo',
                 _version=1.5,
                 _demo=False):
        
        # Create the auth object:
        self.AUTHENTICATION = DWX_API_AUTHENTICATION(_auth_creds)
        
        # Construct main production url for tagging endpoints
        self._url = '{}/{}/{}'.format(_api_url, _api_name, _version)

    def _construct_auth_post_headers(self):

        # Will be called after the new access token is retrieved.

        # Construct authorization header for all requests
        self._auth_headers = {'Authorization': f'Bearer {self.AUTHENTICATION._auth_creds["access_token"]}'}
        
        # Construct headers for POST requests
        self._post_headers = {**self._auth_headers,
                              **{'Content-type':'application/json',
                                 'Accept':'application/json'}}

    @Decorators.refreshTokenDecorator_APIs
    def _Call_API_(self, _endpoint, _type, _data, _json=True, _stream=False):
        
        """Call any endpoint provided in the Darwinex API documentation, and get JSON."""

        # Construct the headers:
        self._construct_auth_post_headers()

        if _type not in ['GET','POST','PUT', 'DELETE']:
            print('Bad request type')
            return None
        
        try:
            
            if _type == 'GET':
                _ret = requests.get(self._url + _endpoint,
                                    headers=self._auth_headers,
                                    verify=True)
                print(f'**** FULL URL ENDPOINT ****: {_ret.url}')
            elif _type == 'PUT':
                _ret = requests.put(self._url + _endpoint,
                                    headers=self._post_headers,
                                    data=_data,
                                    verify=True)
                print(f'**** FULL URL ENDPOINT ****: {_ret.url}')
            elif _type == 'DELETE':
                _ret = requests.delete(self._url + _endpoint,
                                       headers=self._auth_headers,
                                       #data=_data,
                                       verify=True)
                print(f'**** FULL URL ENDPOINT ****: {_ret.url}')
            else:
                if len(_data) == 0:
                    print('Data is empty..')
                    return None
                
                # For DARWIN Quotes API
                if _stream:
                    
                    # Add POST header for streaming quotes        
                    self._post_headers['connection'] = 'keep-alive'
                    return requests.Request('POST',
                                           self._url + _endpoint,
                                           headers=self._post_headers,
                                           data=_data)
                else:
                    _ret = requests.post(self._url + _endpoint,
                                         data=_data,
                                         headers = self._post_headers,
                                         verify=True)
        
            if _json:
                #print(_ret.json())
                return _ret.json()
            else:
                return _ret
        
        except Exception as ex:
            print('Type: {0}, Args: {1!r}'.format(type(ex).__name__, ex.args))
            print(f'Response request code: {_ret.status_code}')
            print(f'Response request URL: {_ret.url}')
            print(f'Response request Data: {_ret.text}')
            
    ##########################################################################



# -*- coding: utf-8 -*-
"""
    DWX Quotes API - Subclass of DWX_API for Quotes Streaming
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: June 14, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""
import os

from darwinexapis.MINIONS.dwx_graphics_helpers import DWX_Graphics_Helpers
from darwinexapis.API.dwx_api import DWX_API

from matplotlib import pyplot as plt
from requests import Session
import pandas as pd
import numpy as np

pd.set_option('display.width', 1000)
pd.set_option('display.max_columns', 500)

class DWX_Quotes_API(DWX_API):
    
    def __init__(self,
                 _auth_creds='',
                 _api_url='https://api.darwinex.com',
                 _api_name='quotes',
                 _version=1.0):
        
        super(DWX_Quotes_API, self).__init__(_auth_creds, _api_url, _api_name, _version)
        self._graphics = DWX_Graphics_Helpers()
        
        # Note: Pass _json=False and _stream=True to self._Call_API_() to 
        # receive request object instead of JSON text output.
    
    ##########################################################################
    
    def _stream_quotes_(self, 
                        _endpoint='/quotes',
                        _symbols=['THA.4.12','LVS.4.20']):
        
        # Create session for streaming quotes
        _s = Session()
        
        # Construct DARWIN list for sending to the endpoint
        _data = "{\"productNames\": [ \"" + "\",\"".join(_symbols) + "\" ]}"
        
        # Retreive Request object for session
        _ret = self._Call_API_(_endpoint,
                               _type='POST',                               
                               _data=_data, 
                               _json=False,
                               _stream=True)
                
        # Send the request
        _resp = _s.send(_ret.prepare(), stream=True, verify=True)
        
        # Yield output
        for _l in _resp.iter_lines():
            if _l:
                yield _l

    ##########################################################################
    
    def _process_stream_(self, 
                         _symbols=['DWZ.4.7','DWC.4.20','LVS.4.20','SYO.4.24','YZZ.4.20'],
                         _plot=True):
       
        # Create empty dataframe
        self._df = pd.DataFrame(columns=_symbols)
        
        # Create graphing objects if _plot==True
        if _plot:
            
            # Create fig, ax
            if len(_symbols) > 1:
                _fig, _ax = plt.subplots(nrows=1, ncols=len(_symbols), figsize=(20,4))
            else:
                _fig, _ax = plt.subplots(figsize=(10,6))        
            
        # for line in self.streaming(symbols=_symbols):
        for _ret in self._stream_quotes_(_symbols=_symbols):
        
            # Parse string to dict
            _stream = eval(_ret)
        
            # Extract values of interest.
            _darwin = _stream['productName']
            _quote = _stream['quote']
            _timestamp = _stream['timestamp']
            
            # Add to DataFrame
            _quotes = [np.nan for x in range(self._df.shape[1])]
            _quotes[_symbols.index(_darwin)] = _quote
            _row = pd.Series(index=_symbols,
                             data=_quotes,
                             name=_timestamp)
            
            self._df = self._df.append(_row)
            self._df.fillna(method='ffill', inplace=True)
            self._df.fillna(method='bfill', inplace=True)
            
            if _plot == False:
                print(_stream)
            else:
                # Axis to pass
                if len(_symbols) > 1:
                    _axp = _ax[_symbols.index(_darwin)]
                else:
                    _axp = _ax
                    
                # Call plotter function from self._graphics
                self._graphics._mpl_plot_axis_(
                        plt, 
                        _axp,
                        #_ax[_symbols.index(_darwin)], 
                        self._df,
                        _darwin,
                        'Last 100 ticks',
                        'Quote',
                        '#00fa9a',
                        0.5,
                        100,
                        '#07335B',
                        {'fontname':'Courier New'})
              
    ##########################################################################



# -*- coding: utf-8 -*-
"""
    DWX_Trading_API.py - DARWIN Trading API (Subclass of DWX_API)
    --
    @author: Darwinex Labs (www.darwinex.com)
    
    Last Updated: July 29, 2019
    
    Copyright (c) 2017-2019, Darwinex. All rights reserved.
    
    Licensed under the BSD 3-Clause License, you may not use this file except 
    in compliance with the License. 
    
    You may obtain a copy of the License at:    
    https://opensource.org/licenses/BSD-3-Clause
"""
import os

from darwinexapis.API.dwx_api import DWX_API

class DWX_Trading_API(DWX_API):
    
    def __init__(self,
                 _auth_creds='',
                 _api_url='https://api.darwinex.com',
                 _api_name='trading',
                 _version=1.0,
                 _demo=True):
        
        super(DWX_Trading_API, self).__init__(_auth_creds, _api_url,_api_name,_version,_demo)
        
        if _demo:
            print(f'--+--+--\n[KERNEL] DARWIN Trading API v{_version} initialized in DEMO environment\n--+--+--')
        else:
            print(f'--+--+--\n[KERNEL] DARWIN Trading API v{_version} initialized in REAL environment\n--+--+--')
        
    #########################################################################
    
    def _Get_Permitted_Operations_(self):
        
        try:
            return self._Call_API_(_endpoint='/productmarket/status',
                                   _type='GET', _data='')
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Get_Account_Leverage_(self, _id=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/leverage',
                                   _type='GET', _data='')
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    # DUMMY ORDER GENERATORS BELOW -- DO NOT USE IN LIVE TRADING PLEASE!
    
    def _generate_dummy_buy_order_(self):
        
        return '{\
          "amount": 200.00,\
          "productName": "NTI.4.12",\
          "thresholdParameters": {\
            "quoteStopLoss": 100.00,\
            "quoteTakeProfit": 1000.00\
          }\
        }'
    
    def _generate_dummy_sell_order_(self):
        
        return '{\
          "amount": 200.00,\
          "productName": "NTI.4.12"\
        }'
        
    def _generate_dummy_buylimit_order_(self):
        
        return '{\
          "amount": 200.00,\
          "productName": "NTI.4.12",\
          "quote": 200.00,\
          "side": "BUY",\
          "type": "LESS_THAN_EQUAL",\
          "thresholdParameters": {\
            "quoteStopLoss": 100.00,\
            "quoteTakeProfit": 1000.00\
          }\
        }'
    
    def _generate_dummy_buystop_order_(self):
        
        return '{\
          "amount": 200.00,\
          "productName": "NTI.4.12",\
          "quote": 500.00,\
          "side": "BUY",\
          "type": "GREATER_THAN_EQUAL",\
          "thresholdParameters": {\
            "quoteStopLoss": 100.00,\
            "quoteTakeProfit": 1000.00\
          }\
        }'
    
    def _generate_dummy_conditional_update_order_(self):
        
        return '{\
          "amount": 200.00,\
          "quote": 150.00,\
          "thresholdParameters": {\
            "quoteStopLoss": 100.00,\
            "quoteTakeProfit": 999.99\
          }\
        }'
    
    # DUMMY ORDER GENERATORS ABOVE -- DO NOT USE IN LIVE TRADING PLEASE!
    
    #########################################################################
    
    def _Raise_Conditional_Order_(self, 
                                  _id=0, # Investor Account ID
                                  _order=None):
        
        # Construct Buy Order dict if arg _order = None
        if _order is None:
            _order = self._generate_dummy_buylimit_order_() # or buystop
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/conditionalorders',
                                   _type = 'POST',
                                   _data = _order)
            
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Update_Conditional_Order_(self, 
                                  _id=0, # Investor Account ID
                                  _oid=0, # Conditional Order ID
                                  _order=None):
        
        # Construct Conditional Order dict if arg _order = None
        if _order is None:
            _order = self._generate_dummy_conditional_update_order_()
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/conditionalorders/'
                                   + str(_oid),
                                   _type = 'PUT',
                                   _data = _order)
            
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Cancel_Conditional_Order_(self, 
                                   _id=0, # Investor Account ID
                                   _oid=0): # Conditional Order ID
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/conditionalorders/'
                                   + str(_oid),
                                   _type = 'DELETE',
                                   _data = '{"0": 0}') # No data required
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Buy_At_Market_(self, 
                        _id=0, # Investor Account ID
                        _order=None):
        
        # Construct Buy Order dict if arg _order = None
        if _order is None:
            _order = self._generate_dummy_buy_order_()
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/orders/buy',
                                   _type = 'POST',
                                   _data = _order)
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
        
    #########################################################################
    
    def _Sell_At_Market(self,
                        _id=0, 
                        _order=None):
        
        # Construct Sell Order dict if arg _order = None
        if _order is None:
            _order = self._generate_dummy_sell_order_()
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/orders/sell',
                                   _type = 'POST',
                                   _data = _order)
            
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
            
    #########################################################################
    
    def _Close_All_Account_Trades_(self, _id=0):
        
        try:
            return self._Call_API_(_endpoint='/investoraccounts/'
                                   + str(_id)
                                   + '/stopout',
                                   _type = 'POST',
                                   _data = '{"0": 0}') # No data required
            
        except Exception as ex:
            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)
    
    #########################################################################
    
    def _Close_All_DARWIN_Trades_(self, 
                                  _id=0,
                                  _darwin='PLF.4.1'):
        
        if _darwin != '':
            try:
                return self._Call_API_(_endpoint='/investoraccounts/'
                                       + str(_id)
                                       + '/stopout/'
                                       + str(_darwin),
                                       _type = 'POST',
                                       _data = '{"0": 0}') # No data required
                
            except Exception as ex:
                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
        else:
            print('[ERROR] No DARWIN Ticker Symbol provided.. please try again.')
    
    #########################################################################



### Do some imports:
import os, pprint, json, base64, requests, time
import pandas as pd

class DWX_API_AUTHENTICATION(object):

    '''This class will handle the neccesary work for refreshing the access
    token. The post response back will be like (access token and refresh token will change):
    
    {'access_token': '63ca7671-cf30-3e4d-a13e-0ae789183d73',
     'expires_in': 3600,
     'id_token': '...eyJ4NXQi...',
     'refresh_token': '827c785e-f5ed-33bc-87e5-fb9854fe1f80',
     'scope': 'openid',
     'token_type': 'Bearer'}'''

    def __init__(self, _auth_creds):

        '''_auth_creds is a dictionary with all the credentials:

        Ex:

        {'access_token': '63ca7671-cf30-3e4d-a13e-0ae789183d73',
         'consumer_key': '827c785e-f5ed-33bc-87e5-fb9854',
         'consumer_secret': '827c785e-f5ed-33bc-87e5-fb9854',
         'refresh_token': '827c785e-f5ed-33bc-87e5-fb9854fe1f80'}'''
 
        self._auth_creds = _auth_creds

        # Put 60 seconds to request the new one and start from there:
        self.expires_in = time.time() + 60
        print(f'[INIT] - Access token will be created again at {self.expires_in} UNIX timestamp')
 
    def _get_access_refresh_tokens_wrapper(self):

        # Wrap the get tokens method to be called:
        try:
            self.access_token, self.expires_in, self.refresh_token = self._get_access_refresh_tokens_(self._auth_creds['consumer_key'], 
                                                                                                      self._auth_creds['consumer_secret'], 
                                                                                                      self._auth_creds['refresh_token'])

            # The access token is None > Raise exception.
            if self.access_token is None:

                raise Exception("[INIT] - Request for access token failed > No access_token returned in the response")

            #print(f'[REFRESH] - New access_token > {self.access_token}')
            #print(f'[REFRESH] - New expires_in > {self.expires_in}')
            #print(f'[REFRESH] - New refresh_token > {self.refresh_token}')

            print("[INIT] - Will sleep for some secs...")
            time.sleep(3)

        except Exception as ex:
            print(ex)

        else:
            # It is useful for code that must be executed if the try clause does not raise an exception.
            print("[REFRESH_ELSE] - Will RESET the expires_in variable...")
            self.expires_in = time.time() + (self.expires_in - 300)

            # Put new credentials:
            print("[REFRESH_ELSE] - Will add the new access_token and refresh_token to the dictionary")
            self._auth_creds['access_token'] = self.access_token
            self._auth_creds['refresh_token'] = self.refresh_token
            #print(f"[REFRESH_ELSE] - NEW CREDS: {self._auth_creds}")

    def _get_access_refresh_tokens_(self, client_id, client_secret, refresh_token, token_url='https://api.darwinex.com/token'):

        # Refs: https://auth0.com/docs/tokens/guides/use-refresh-tokens
        # Refs: https://help.darwinex.com/api-walkthrough
        # Refs: https://help.darwinex.com/
    
        header_data = {'client_id': client_id, # Consumer key in the Darwinex web
                       'client_secret': client_secret} # Consumer secret in the Darwinex web

        data = {'grant_type': 'refresh_token',
                'refresh_token': refresh_token}
    
        headers = {'Authorization': 'Basic {}'.format(base64.b64encode(bytes('{}:{}'.format(header_data['client_id'], header_data['client_secret']).encode('utf-8'))).decode('utf-8'))}

        try:
            _response = requests.post(token_url, headers=headers, data=data, verify=True, allow_redirects=False)
            
            # Optional: Raise an exception if a request is unsuccessful:
            _response.raise_for_status()

            print('[GET_TOKENS] - Access & Refresh Tokens Retrieved Successfully')
            
            response_json = json.loads(_response.text)
            #print(f'[GET_TOKENS] - RETURNED RESPONSE IN TEXT: {response_json}')
            return response_json['access_token'], response_json['expires_in'], response_json['refresh_token']
            
        except Exception as ex:

            print('[GET_TOKENS] - Access & Refresh Tokens ***NOT*** Retrieved Successfully')
            print('Type: {0}, Args: {1!r}'.format(type(ex).__name__, ex.args))
            return None, None, None



"""Darwinex APIs setup script."""

import sys
from pathlib import Path

from setuptools import setup, find_packages

here = Path(__file__).parent.resolve()

__version__ = None
with open(here / 'version.py') as f:
    exec(f.read())
print(here)

with open(here / 'README.rst', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='darwinexapis',
    version=__version__,
    description='Python wrapper for Darwinex APIs',
    long_description=long_description,
    url='https://github.com/darwinex/darwinexapis',
    author='Alpha Team @ DWX',
    author_email='content@darwinex.com',
    license='BSD',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Office/Business :: Financial :: Investment',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    platforms = ['any'],
    keywords='darwinex algorithmic trading quant quantitative analysis asyncio websockets darwins async',
    #package_dir={"": "darwinexapis"},
    include_package_data=True,
    packages=find_packages(),
    #package_data={'darwinexapis':['*']},
    install_requires=['pandas',
                      'plotly', 
                      'matplotlib', 
                      'requests', 
                      'websockets',
                      'tqdm',
                      'dash',
                      'pyftpdlib',
                      'tables',
                      'matplotlib'
                      ]
)




import json, base64, requests, pprint

class DWX_OAuth2():
    
    def __init__(self):
        
        pass
    
    ##########################################################################

    def _get_tokens_(self, token_url='https://api.darwinex.com/token'):

        # Refs: https://auth0.com/docs/tokens/guides/use-refresh-tokens
        # https://help.darwinex.com/api-walkthrough
        # https://help.darwinex.com/
    
        header_data = {'client_id': '', ## Consumer key in the Darwinex web
                       'client_secret': ''} ## Consumer secret in the Darwinex web

        data = {'grant_type': 'refresh_token',
                'refresh_token': ''}
    
        headers = {'Authorization': 'Basic {}'.format(base64.b64encode(bytes('{}:{}'.format(header_data['client_id'], header_data['client_secret']).encode('utf-8'))).decode('utf-8'))}

        try:
            _response = requests.post(token_url, headers=headers, data=data, verify=True, allow_redirects=False)
            
            print('[KERNEL] Access & Refresh Tokens Retrieved Successfully')
            
            response_json = json.loads(_response.text)
            #return response_json
            return response_json['access_token'], response_json['expires_in'], response_json['refresh_token']
            
        except Exception as ex:
            print('Type: {0}, Args: {1!r}'.format(type(ex).__name__, ex.args))
            return None
    
    ##########################################################################

if __name__ == "__main__":

    '''Example response:
    
    {'access_token': '63ca7671-cf30-3e4d-a13e-0ae78781183d73',
     'expires_in': 3600,
     'id_token': '...',
     'refresh_token': '827c785e-f5ed-33bc-87e5-fb91354fe1f80',
     'scope': 'openid',
     'token_type': 'Bearer'}'''

    a = DWX_OAuth2()
    tokens = a._get_tokens_()

    pprint.pprint(tokens)


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 29 17:24:20 2018
Script: dwx_tickdata_download.py (Python 3)
--
Downloads tick data from the Darwinex tick data server. This code demonstrates
how to download data for one specific date/hour combination, but can be 
extended easily to downloading entire assets over user-specified start/end 
datetime ranges.
Requirements: Your Darwinex FTP credentials.
Result: Dictionary of pandas DataFrame objects by date/hour.
        (columns: float([ask, size]), index: millisecond timestamp)
        
Example code:
    > td = DWX_Tick_Data(dwx_ftp_user='very_secure_username', 
                         dwx_ftp_pass='extremely_secure_password',
                         dwx_ftp_hostname='mystery_ftp.server.com', 
                         dwx_ftp_port=21)
    
    > td._download_one_hour_data_ask(_asset='EURNOK', _date='2018-10-22', _hour='00')
    
    > td._asset_db['EURNOK-2018-10-22-00']
    
                                           ask       size
     2018-10-22 00:00:07.097000+00:00  9.47202  1000000.0
     2018-10-22 00:00:07.449000+00:00  9.47188   750000.0
     2018-10-22 00:01:08.123000+00:00  9.47201   250000.0
     2018-10-22 00:01:10.576000+00:00  9.47202  1000000.0
                                  ...        ...
@author: Darwinex Labs
@twitter: https://twitter.com/darwinexlabs
@web: http://blog.darwinex.com/category/labs
"""

# Do some imports:
from ftplib import FTP 
from io import BytesIO
import pandas as pd
import gzip, os
from datetime import timedelta, date, datetime

class DWX_TickData_Downloader_API(object):
    
    def __init__(self, dwx_ftp_user, dwx_ftp_pass, dwx_ftp_hostname, dwx_ftp_port):
        
        # Dictionary DB to hold dictionary objects in FX/Hour format
        self._asset_db = {}
        
        self._ftpObj = FTP(dwx_ftp_hostname)                            
        self._ftpObj.login(dwx_ftp_user, dwx_ftp_pass)   

        self._virtual_dl = None
        
    ############################## DOWNLOAD DATA ##############################
    
    def _download_one_hour_data_ask(self, _asset='EURUSD', _date='2017-10-01', _hour='22', _ftp_loc_format='{}/{}_ASK_{}_{}.log.gz', _verbose=False):

        # Function: Downloads and stored currency tick data from Darwinex FTP
        #           Server. Object stores data in a dictionary, keys being of the
        #           format: CURRENCYPAIR-YYYY-MM-DD-HH
        
        _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
        self._file_name_csv = f'{_asset}_ASK_{_date}_{_hour}.csv'
        _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
        self._virtual_dl = BytesIO()
        
        if _verbose is True:
            print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
        try:
            self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)

            self._virtual_dl.seek(0)
            _log = gzip.open(self._virtual_dl)
                
            # Get bytes into local DB as list of lists
            self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
            # Construct DataFrame
            _temp = self._asset_db[_key]
            self._asset_db[_key] = pd.DataFrame({'ask': [l[1] for l in _temp],
                                'size': [l[2] for l in _temp]}, 
                                index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
            # Sanitize types
            self._asset_db[_key] = self._asset_db[_key].astype(float)
            
            if _verbose is True:
                print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

            # Return the data:
            return self._return_pandas_dataframe_()
        
        # Case: if file not found
        except Exception as ex:

            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)

    def _download_one_hour_data_bid(self, _asset='EURUSD', _date='2017-10-01', _hour='22', _ftp_loc_format='{}/{}_BID_{}_{}.log.gz', _verbose=False):

        # Function: Downloads and stored currency tick data from Darwinex FTP
        #           Server. Object stores data in a dictionary, keys being of the
        #           format: CURRENCYPAIR-YYYY-MM-DD-HH
        
        _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
        self._file_name_csv = f'{_asset}_BID_{_date}_{_hour}.csv'
        _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
        self._virtual_dl = BytesIO()
        
        if _verbose is True:
            print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
        try:
            self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)
                
            self._virtual_dl.seek(0)
            _log = gzip.open(self._virtual_dl)
                
            # Get bytes into local DB as list of lists
            self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
            # Construct DataFrame
            _temp = self._asset_db[_key]
            self._asset_db[_key] = pd.DataFrame({'bid': [l[1] for l in _temp],
                                'size': [l[2] for l in _temp]}, 
                                index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
            # Sanitize types
            self._asset_db[_key] = self._asset_db[_key].astype(float)
            
            if _verbose is True:
                print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

            # Return the data:
            return self._return_pandas_dataframe_()
        
        # Case: if file not found
        except Exception as ex:

            _exstr = "Exception Type {0}. Args:\n{1!r}"
            _msg = _exstr.format(type(ex).__name__, ex.args)
            print(_msg)

    ##########################################################

    def _download_one_day_data_ask(self, _asset='EURUSD', _date='2017-10-01', _ftp_loc_format='{}/{}_ASK_{}_{}.log.gz', _verbose=False):

        # This method will download data for one day, all hours:
        
        full_dataframe = pd.DataFrame(columns=['ask', 'size'])

        # You can change the range parameter to alter the hours of download.
        # The final part of the csv name will have the last hour of data that was requested.
        for _hour in range(0, 24):

            _hour = str(_hour).zfill(2)

            _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
            self._file_name_csv = f'{_asset}_ASK_{_date}_{_hour}.csv'
            _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
            self._virtual_dl = BytesIO()
        
            if _verbose is True:
                print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
            try:
                self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)
                
                self._virtual_dl.seek(0)
                _log = gzip.open(self._virtual_dl)
                
                # Get bytes into local DB as list of lists
                self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
                # Construct DataFrame
                _temp = self._asset_db[_key]
                self._asset_db[_key] = pd.DataFrame({'ask': [l[1] for l in _temp],
                                    'size': [l[2] for l in _temp]}, 
                                    index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
                # Sanitize types
                self._asset_db[_key] = self._asset_db[_key].astype(float)
            
                if _verbose is True:
                    print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

                # Return the data:
                hour_df_data = self._return_pandas_dataframe_()
                self._asset_db = {}

                # Concat the data:
                full_dataframe = full_dataframe.append(hour_df_data)
        
            # Case: if file not found
            except Exception as ex:

                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
                print('[ERROR] - Check if the date provided is a valid trading day')

        # Return the data finally:
        self._asset_db = {}
        return full_dataframe

    def _download_one_day_data_bid(self, _asset='EURUSD', _date='2017-10-01', _ftp_loc_format='{}/{}_BID_{}_{}.log.gz', _verbose=False):

        # This method will download data for one day, all hours:
        
        full_dataframe = pd.DataFrame(columns=['bid', 'size'])

        # You can change the range parameter to alter the hours of download.
        # The final part of the csv name will have the last hour of data that was requested.
        for _hour in range(0, 24):

            _hour = str(_hour).zfill(2)

            _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
            self._file_name_csv = f'{_asset}_BID_{_date}_{_hour}.csv'
            _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
            self._virtual_dl = BytesIO()
        
            if _verbose is True:
                print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
            try:
                self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)
                
                self._virtual_dl.seek(0)
                _log = gzip.open(self._virtual_dl)
                
                # Get bytes into local DB as list of lists
                self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
                # Construct DataFrame
                _temp = self._asset_db[_key]
                self._asset_db[_key] = pd.DataFrame({'bid': [l[1] for l in _temp],
                                    'size': [l[2] for l in _temp]}, 
                                    index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
                # Sanitize types
                self._asset_db[_key] = self._asset_db[_key].astype(float)
            
                if _verbose is True:
                    print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

                # Return the data:
                hour_df_data = self._return_pandas_dataframe_()
                self._asset_db = {}

                # Concat the data:
                full_dataframe = full_dataframe.append(hour_df_data)
        
            # Case: if file not found
            except Exception as ex:

                _exstr = "Exception Type {0}. Args:\n{1!r}"
                _msg = _exstr.format(type(ex).__name__, ex.args)
                print(_msg)
                print('[ERROR] - Check if the date provided is a valid trading day')

        # Return the data finally:
        self._asset_db = {}
        return full_dataframe

    ##########################################################

    def date_range(self, start_date, end_date):

        # Generator method to yield the following date:

        # Parse the entry strings to date:
        start_date = datetime.strptime(start_date, "%Y-%m-%d").date()
        end_date = datetime.strptime(end_date, "%Y-%m-%d").date()

        for each_day in range(int((end_date - start_date).days)):
            yield start_date + timedelta(each_day)

    def _download_month_data_ask(self, _asset='EURUSD', _start_date='2017-10-01', _end_date='2017-11-01', 
                                 _ftp_loc_format='{}/{}_ASK_{}_{}.log.gz', _verbose=False):

        # This method will download data for one month, all days available and all hours:
        
        full_dataframe = pd.DataFrame(columns=['ask', 'size'])

        # NOTE: For consistency with the built-in range() function this iteration stops before reaching the end_date. 
        # So for inclusive iteration use the next day, as you would with range()

        for _date in self.date_range(_start_date, _end_date):

            # Parse the date to string:
            _date = _date.strftime("%Y-%m-%d")

            # You can change the range parameter to alter the hours of download.
            # The final part of the csv name will have the last hour of data that was requested and last date.
            for _hour in range(0, 24):

                _hour = str(_hour).zfill(2)

                _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
                self._file_name_csv = f'{_asset}_ASK_{_date}_{_hour}.csv'
                _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
                self._virtual_dl = BytesIO()
        
                if _verbose is True:
                    print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
                try:
                    self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)
                
                    self._virtual_dl.seek(0)
                    _log = gzip.open(self._virtual_dl)
                
                    # Get bytes into local DB as list of lists
                    self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
                    # Construct DataFrame
                    _temp = self._asset_db[_key]
                    self._asset_db[_key] = pd.DataFrame({'ask': [l[1] for l in _temp],
                                        'size': [l[2] for l in _temp]}, 
                                        index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
                    # Sanitize types
                    self._asset_db[_key] = self._asset_db[_key].astype(float)
            
                    if _verbose is True:
                        print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

                    # Return the data:
                    hour_df_data = self._return_pandas_dataframe_()
                    self._asset_db = {}

                    # Concat the data:
                    full_dataframe = full_dataframe.append(hour_df_data)
        
                # Case: if file not found
                except Exception as ex:

                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
                    print('[ERROR] - Check if the date provided is a valid trading day')

        # Return the data finally:
        self._asset_db = {}
        return full_dataframe

    def _download_month_data_bid(self, _asset='EURUSD', _start_date='2017-10-01', _end_date='2017-11-01', 
                                 _ftp_loc_format='{}/{}_BID_{}_{}.log.gz', _verbose=False):

        # This method will download data for one month, all days available and all hours:
        
        full_dataframe = pd.DataFrame(columns=['bid', 'size'])

        # NOTE: For consistency with the built-in range() function this iteration stops before reaching the end_date. 
        # So for inclusive iteration use the next day, as you would with range()

        for _date in self.date_range(_start_date, _end_date):

            # Parse the date to string:
            _date = _date.strftime("%Y-%m-%d")

            # You can change the range parameter to alter the hours of download.
            # The final part of the csv name will have the last hour of data that was requested and last date.
            for _hour in range(0, 24):

                _hour = str(_hour).zfill(2)

                _file = _ftp_loc_format.format(_asset, _asset, _date, _hour)
                self._file_name_csv = f'{_asset}_BID_{_date}_{_hour}.csv'
                _key = '{}-{}-{}'.format(_asset, _date, _hour)
        
                self._virtual_dl = BytesIO()
        
                if _verbose is True:
                    print("\n[INFO] Retrieving file \'{}\' from Darwinex Tick Data Server..".format(_file))
        
                try:
                    self._ftpObj.retrbinary("RETR {}".format(_file), self._virtual_dl.write)
                
                    self._virtual_dl.seek(0)
                    _log = gzip.open(self._virtual_dl)
                
                    # Get bytes into local DB as list of lists
                    self._asset_db[_key] = [line.strip().decode().split(',') for line in _log]
            
                    # Construct DataFrame
                    _temp = self._asset_db[_key]
                    self._asset_db[_key] = pd.DataFrame({'bid': [l[1] for l in _temp],
                                        'size': [l[2] for l in _temp]}, 
                                        index=[pd.to_datetime(l[0], unit='ms', utc=True) for l in _temp])
            
                    # Sanitize types
                    self._asset_db[_key] = self._asset_db[_key].astype(float)
            
                    if _verbose is True:
                        print('\n[SUCCESS] {} tick data for {} (hour {}) stored in self._asset_db dict object.\n'.format(_asset, _date, _hour))

                    # Return the data:
                    hour_df_data = self._return_pandas_dataframe_()
                    self._asset_db = {}

                    # Concat the data:
                    full_dataframe = full_dataframe.append(hour_df_data)
        
                # Case: if file not found
                except Exception as ex:

                    _exstr = "Exception Type {0}. Args:\n{1!r}"
                    _msg = _exstr.format(type(ex).__name__, ex.args)
                    print(_msg)
                    print('[ERROR] - Check if the date provided is a valid trading day')

        # Return the data finally:
        self._asset_db = {}
        return full_dataframe

    ##########################################################

    def _return_pandas_dataframe_(self):

        # Transform the data to a pandas dataframe:
        dataframe_values = self._asset_db.get(list(self._asset_db.keys())[0])

        # Return it:
        self._asset_db = {}
        return dataframe_values

    ##########################################################

    def _save_df_to_csv(self, dataframe_to_save, which_path=None):

        # Save:
        if which_path:
            # It will save the data to the specified path:
            dataframe_to_save.to_csv(which_path + self._file_name_csv)
        else:
            # It will save the data in the working directory:
            dataframe_to_save.to_csv(self._file_name_csv)

    def _save_df_to_pickle(self, dataframe_to_save, which_path=None):

        # Save:
        if which_path:
            # It will save the data to the specified path:
            dataframe_to_save.to_pickle(which_path + self._file_name_csv.split('.')[0] + '.pkl')
        else:
            # It will save the data in the working directory:
            dataframe_to_save.to_pickle(self._file_name_csv.split('.')[0] + '.pkl')

    def _save_df_to_hdf(self, dataframe_to_save, which_path=None):

        # Save:
        if which_path:
            # It will save the data to the specified path:
            dataframe_to_save.to_hdf(which_path + self._file_name_csv.split('.')[0] + '.h5', key='df', mode='w')
        else:
            # It will save the data in the working directory:
            dataframe_to_save.to_csv(self._file_name_csv.split('.')[0] + '.h5')


### Do some imports:
import os, pprint
import pandas as pd

### Import the different classes:
from darwinexapis.API.InfoAPI.DWX_Info_API import DWX_Info_API
from darwinexapis.API.InvestorAccountInfoAPI.DWX_AccInfo_API import DWX_AccInfo_API
from darwinexapis.API.QuotesAPI.DWX_Quotes_API import DWX_Quotes_API
from darwinexapis.API.TradingAPI.DWX_Trading_API import DWX_Trading_API
from darwinexapis.API.WebSocketAPI.DWX_WebSocket_API import DWX_WebSocket_API

### Let's create the access token variable:
AUTH_CREDS = {'access_token': 'YOUR_ALPHA_TOKEN',
              'consumer_key': 'YOUR_ALPHA_TOKEN',
              'consumer_secret': 'YOUR_ALPHA_TOKEN',
              'refresh_token': 'YOUR_ALPHA_TOKEN'}

### Let's instantiate some objects:
#darwinexInfo = DWX_Info_API(AUTH_CREDS, _version=2.0, _demo=True)
#darwinexInvestorAcc = DWX_AccInfo_API(AUTH_CREDS, _version=2.0, _demo=True)
#darwinexQuotes = DWX_Quotes_API(AUTH_CREDS, _version=1.0)
#darwinexTrading = DWX_Trading_API(AUTH_CREDS, _version=1.1, _demo=True)
darwinexWebSocket = DWX_WebSocket_API(AUTH_CREDS, _version=0.0)

### DWX_Info_API:
#darwinUniverse = darwinexInfo._Get_DARWIN_Universe_(_status='ACTIVE', _iterate=True, _perPage=50)
#pprint.pprint(darwinUniverse)
#darwinUniverse.to_csv('/home/zmlaptop/Desktop/AlphaTeamDX/DataDARWINUniverse/darwinUniverse.csv')

### DWX_AccInfo_API:
#print(darwinexInvestorAcc._Get_Accounts_())

### DWX_Quotes_API:
#darwinexQuotes._stream_quotes_()
#darwinexQuotes._process_stream_(_symbols=["ENH.4.16"], _plot=False)

### DWX_Trading_API:
#darwinexTrading._Get_Permitted_Operations_()
#darwinexTrading._Get_Account_Leverage_(_id=0)

### DWX_WebSocket_API:
darwinexWebSocket.run(_symbols=["ENH.4.16", "CIS.4.11", "CGT.4.5","CDG.4.14", "ABH.4.21", "ENO.4.13"])